
#include <string>
#include <set>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdlib>
#include <ctime>
#include <map>
#include <stdexcept>
#include <deque>

#include "hash.hpp"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

std::deque<std::string> directory_xini_files(std::string dir)
{
	std::deque<std::string> result;
	
	WIN32_FIND_DATA ffd;
	HANDLE ffh = FindFirstFile((dir + "\\*.xini").c_str(), &ffd);
	
	if (ffh == INVALID_HANDLE_VALUE)
		throw std::runtime_error("Directory not found: " + dir);
	
	do
	{
		if (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			continue;
		
		result.push_back(dir + "\\" + ffd.cFileName);

	} while (FindNextFile(ffh, &ffd) != 0);
	
	FindClose(ffh);
	
	return result;
}

std::string X(std::string in)
{
	std::string result;
	
	for (char c : in)
	{
		if (c == '\'' || c == '\\')
			result += '\\';

		result += c;
	}
	
	return result;
}

template <class T> T from_string(std::string x)
{
	std::stringstream ss(x);
	T result;
	ss >> result;
	return result;
}

template <class T> std::string to_string(const T& x)
{
	std::stringstream ss;
	ss << x;
	return ss.str();
}

std::string inventory_convert(std::string string)
{
	for (char& c : string)
	{
		if (c == ':')
			c = ',';
		else if (c == ',')
			c = ';';
	}

	return string;
}

struct seose_character
{
	std::string name;
	std::string title;
	int gender;
	int hairstyle;
	int haircolor;
	std::string partner;
	int race;
	int admin;
	int class_;
	int map;
	int x;
	int y;
	int direction;
	int sitting;
	int usage;
	int exp;
	int hp;
	int tp;
	int karma;
	int statpoints;
	int skillpoints;
	int level;
	int str;
	int int_;
	int wis;
	int agi;
	int con;
	int cha;
	std::string paperdoll;
	std::string inventory;
	std::string bank;
	std::string spells;
	int bankmax;
	int goldbank;
		
	bool operator <(const seose_character& rhs) const
	{
		return name < rhs.name;
	}
};

struct seose_account
{
	std::string name;
	std::string password;
	std::string fullname;
	std::string location;
	std::string email;
	std::string computer;
	int hdid;
	std::string ip;
	
	bool operator <(const seose_account& rhs) const
	{
		return name < rhs.name;
	}
};

std::map<std::string, std::string> account_map;

seose_account parse_account_file(std::string filename)
{
	std::size_t dir_sep_pos = filename.find_first_of('\\');
	
	if (dir_sep_pos == std::string::npos)
		dir_sep_pos = 0;
	else
		++dir_sep_pos;

	std::string basename = filename.substr(dir_sep_pos);
	std::string name = basename.substr(0, basename.length() - 5);
	seose_account a{name};
	std::ifstream is(filename);
	
	std::string s;
	std::string section;
	
	while (std::getline(is, s))
	{
		if (s[0] == '[')
		{
			std::size_t last_b = s.find_first_of(']');
			section = s.substr(1, last_b - 1);
			
			if (section == "character")
			{
				std::string char_name = s.substr(last_b + 2);
				account_map[char_name] = name;
			}
			
			continue;
		}

		std::size_t first_eq = s.find_first_of('=');
		
		if (first_eq == std::string::npos)
			continue;
		
		std::size_t first_q = s.find_first_of('"');
		std::size_t last_q = s.find_last_of('"');
		
		std::string key = s.substr(0, first_eq);
		std::string data = s.substr(first_q + 1, last_q - first_q - 1);
		
		     if (key == "password") a.password = data;
		else if (key == "fullname") a.fullname = data;
		else if (key == "location") a.location = data;
		else if (key == "emailaddress") a.email = data;
		else if (key == "computername") a.computer = data;
		else if (key == "hddserial") a.hdid = from_string<unsigned long>(data);
		else if (key == "ip") a.ip = data;
		
	}
	
	return a;
}

seose_character parse_character_file(std::string filename)
{
	std::size_t dir_sep_pos = filename.find_first_of('\\');
	
	if (dir_sep_pos == std::string::npos)
		dir_sep_pos = 0;
	else
		++dir_sep_pos;

	std::string basename = filename.substr(dir_sep_pos);
	std::string name = basename.substr(0, basename.length() - 5);
	seose_character c{name};
	std::ifstream is(filename);
	
	std::string s;
	std::string section;
	
	while (std::getline(is, s))
	{
		if (s[0] == '[')
		{
			std::size_t last_b = s.find_first_of(']');
			section = s.substr(1, last_b - 1);			
			continue;
		}

		std::size_t first_eq = s.find_first_of('=');
		
		if (first_eq == std::string::npos)
			continue;
		
		std::size_t first_q = s.find_first_of('"');
		std::size_t last_q = s.find_last_of('"');
		
		std::string key = s.substr(0, first_eq);
		std::string data = s.substr(first_q + 1, last_q - first_q - 1);
		
		if (section == "character")
		{
			      if (key == "sex") c.gender = from_string<int>(data);
			 else if (key == "hairstyle") c.hairstyle = from_string<int>(data);
			 else if (key == "haircolour") c.haircolor = from_string<int>(data);
			 else if (key == "race") c.race = from_string<int>(data);
			 else if (key == "admin") c.admin = from_string<int>(data);
			 else if (key == "classid") c.class_ = from_string<int>(data);
			 else if (key == "title") c.title = data;
			 else if (key == "partner") c.partner = data;
		}
		else if (section == "location")
		{
			     if (key == "map") c.map = from_string<int>(data);
			else if (key == "x") c.x = from_string<int>(data);
			else if (key == "y") c.y = from_string<int>(data);
			else if (key == "d") c.direction = from_string<int>(data);
			else if (key == "sitting") c.sitting = from_string<int>(data);
		}
		else if (section == "stats")
		{
			     if (key == "usage") c.usage = from_string<int>(data);
			else if (key == "exp") c.exp = from_string<int>(data);
			else if (key == "hp") c.hp = from_string<int>(data);
			else if (key == "tp") c.tp = from_string<int>(data);
			else if (key == "karma") c.karma = from_string<int>(data);
			else if (key == "statpoints") c.statpoints = from_string<int>(data);
			else if (key == "skillpoints") c.skillpoints = from_string<int>(data);
		}
		else if (section == "base")
		{
			     if (key == "level") c.level = from_string<int>(data);
			else if (key == "str") c.str = from_string<int>(data);
			else if (key == "int") c.int_ = from_string<int>(data);
			else if (key == "wis") c.wis = from_string<int>(data);
			else if (key == "agi") c.agi = from_string<int>(data);
			else if (key == "con") c.con = from_string<int>(data);
			else if (key == "cha") c.cha = from_string<int>(data);
		}
		else if (section == "items")
		{
			     if (key == "paperdoll") c.paperdoll = data;
			else if (key == "inventory") c.inventory = inventory_convert(data);
			else if (key == "locker") c.bank = inventory_convert(data);
			else if (key == "lockerupgrades") c.bankmax = from_string<int>(data);
			else if (key == "bankedgold") c.goldbank = from_string<int>(data);
		}
		else if (section == "spells")
		{
			     if (key == "spells") c.spells = inventory_convert(data);
		}
	}
	
	return c;
}

unsigned long now;
std::string password_key;

std::set<seose_account> accounts;
std::set<seose_character> characters;

seose_account find_account(std::string name)
{
	auto result = accounts.find(seose_account{name});
	
	if (result == accounts.end())
		throw std::runtime_error("Account not found: " + name);
	
	return *result;
}

seose_character find_character(std::string name)
{
	auto result = characters.find(seose_character{name});
	
	if (result == characters.end())
		throw std::runtime_error("Character not found: " + name);
	
	return *result;
}

seose_account account_for_character(seose_character c)
{
	auto it = account_map.find(c.name);
	
	if (it == account_map.end())
		throw std::runtime_error("Account not found for character: " + c.name);
	
	return find_account(it->second);
}

std::string to_sql(const seose_account& a)
{
	return "INSERT INTO accounts ("
		"username, password, fullname, location, "
		"email, computer, hdid, regip, lastip, "
		"created, lastused"
	") VALUES ("
		"'" + X(a.name) + "', '" + X(sha256(password_key + a.name + a.password)) + "', '" + X(a.fullname) + "', '" + X(a.location) + "', "
		+ "'" + X(a.email) + "', '" + X(a.computer) + "', " + to_string(a.hdid) + ", '" + X(a.ip) + "', '" + X(a.ip) + "', "
		+ to_string(now) + ", " + to_string(now)
	+ ")";
}

std::string to_sql(const seose_character& c)
{
	std::string acc_name = account_map[c.name];

	return "INSERT INTO characters ("
		"name, account, title, "
		"partner, admin, class, gender, race, "
		"hairstyle, haircolor, map, x, y, "
		"direction, level, exp, hp, tp, "
		"str, `int`, wis, agi, con, cha, "
		"statpoints, skillpoints, karma, "
		"sitting, bankmax, goldbank, `usage`, "
		"inventory, bank, paperdoll, spells, "
		"quest, vars"
	") VALUES ("
		"'" + X(c.name) + "', '" + X(acc_name) + "', '" + X(c.title) + "', "
		+ "'" + X(c.partner) + "', " + to_string(c.admin) + ", " + to_string(c.class_) + ", " + to_string(c.gender) + ", " + to_string(c.race) + ", "
		+ to_string(c.hairstyle) + ", " + to_string(c.haircolor) + ", " + to_string(c.map) + ", " + to_string(c.x) + ", " + to_string(c.y) + ", "
		+ to_string(c.direction) + ", " + to_string(c.level) + ", " + to_string(c.exp) + ", " + to_string(c.hp) + ", " + to_string(c.tp) + ", "
		+ to_string(c.str) + ", " + to_string(c.int_) + ", " + to_string(c.wis) + ", " + to_string(c.agi) + ", " + to_string(c.con) + ", " + to_string(c.cha) + ", "
		+ to_string(c.statpoints) + ", " + to_string(c.skillpoints) + ", " + to_string(c.karma) + ", " + to_string(c.sitting) + ", " + to_string(c.bankmax) + ", "
		+ to_string(c.goldbank) + ", " + to_string(c.usage) + ", '" + X(c.inventory) + "', '" + X(c.bank) + "', '" + X(c.paperdoll) + "', '" + X(c.spells) + "', "
		+ "'', ''"
	+ ")";
}

int main()
{
	try
	{
		password_key = "ChangeMe";
		
		std::ofstream account_sql_out("accounts.sql");
		std::ofstream character_sql_out("characters.sql");

		now = time(nullptr);
		now -= now % 1800;
		
		std::cout << "Enumerating account files..." << std::endl;
		std::deque<std::string> account_files = directory_xini_files("Users/Accounts");
		
		std::cout << "Enumerating character files..." << std::endl;
		std::deque<std::string> character_files = directory_xini_files("Users/Characters");
		
		std::cout << "Parsing account files (" << to_string(account_files.size()) << ")..." << std::endl;
		for (std::string s : account_files)
			accounts.insert(parse_account_file(s));
		
		std::cout << "Parsing character files (" << to_string(character_files.size()) << ")..." << std::endl;
		for (std::string s : character_files)
			characters.insert(parse_character_file(s));
		
		std::cout << "Generating account SQL..." << std::endl;
		for (const seose_account& a : accounts)
		{
			account_sql_out << to_sql(a) << ';' << std::endl;
		}
		
		std::cout << "Generating character SQL..." << std::endl;
		for (const seose_character& c : characters)
		{
			character_sql_out << to_sql(c) << ';' << std::endl;
		}
		
		std::cout << "Done!" << std::endl;
	}
	catch (std::runtime_error& e)
	{
		std::cerr << e.what();
		std::ofstream error_out("s2e_error.txt");
		error_out << e.what();
		error_out.close();
		std::terminate();
	}
}
